# Generated by Django 2.2.12 on 2020-04-24 05:14

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('mellon', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='SessionIndex',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('session_index', models.TextField(verbose_name='SAML SessionIndex')),
                ('session_key', models.CharField(max_length=40, verbose_name='Django session key')),
                ('saml_identifier', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='mellon.UserSAMLIdentifier', verbose_name='SAML identifier')),
            ],
            options={
                'verbose_name': 'SAML SessionIndex',
                'verbose_name_plural': 'SAML SessionIndexes',
                'unique_together': {('saml_identifier', 'session_index', 'session_key')},
            },
        ),
    ]
